@@ -85,6 +85,11 @@ skip_enh_tibble_null <- function() {
   skip_legacy()
 }
 
+skip_enh_new_tibble_nrow_null <- function() {
+  # ENH: new_tibble(nrow = NULL), #781
+  skip_legacy()
+}
+
 skip_enh_empty_tribble_unspecified <- function() {
   # ENH: zero-row tribbles create unspecified columns
   skip_legacy()
@@ -105,6 +110,11 @@ skip_enh_enframe_vector <- function() {
   skip_legacy()
 }
 
+skip_enh_print_tbl_args <- function() {
+  # ENH: print() and format() support more arguments
+  skip_legacy()
+}
+
 skip_int_error_unknown_names <- function() {
   # INT: error_unknown_names() no longer implemented
   skip_legacy()
